//    WeIRCd administration tool
//    Copyright (C) 2009 Jonas Thiem
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License v3 as published
//    by the Free Software Foundation.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

// This file provides a small collection of functions for handling linked lists

#include "admintool.h"

#define NONATIVEWEIRCD

#ifndef NONATIVEWEIRCD
#include "weircd.h"
#endif

int LLISTDEBUG = 0;

struct linkedlistitem;
struct linkedlistitem {
	void* data;
	struct linkedlistitem* next;
	struct linkedlistitem* prev;
};

void* linkedlist_Add(void** ptr, const void* item, int itemsize) {
	struct linkedlistitem* list = (struct linkedlistitem*)*ptr;
	if (list == NULL) {
		list = (struct linkedlistitem*)malloc(sizeof(struct linkedlistitem));
		if (list == NULL) {
			return NULL;
		}
		memset(list,0,sizeof(struct linkedlistitem));
		list->data = malloc(itemsize);
		if (list->data == NULL) {
			free(list);
			return NULL;
		}
		memcpy(list->data,item,itemsize);
		*ptr = (void*)list;
		return (void*)list;
	}else{
		struct linkedlistitem* oldnext = list->next;
		list->next = (struct linkedlistitem*)malloc(sizeof(struct linkedlistitem));
		if (list->next == NULL) {
			return NULL;
		}
		memset(list->next,0,sizeof(struct linkedlistitem));
		list->next->prev = list;
		list->next->data = malloc(itemsize);
		if (list->next->data == NULL) {
			free(list->next);
			list->next = oldnext;
			return NULL;
		}
		memcpy(list->next->data,item,itemsize);
		list->next->next = oldnext;
		if (oldnext != NULL) {
			oldnext->prev = list->next;
		}
		return list->next;
	}
}
void* linkedlist_Append(void** ptr, const void* item, int itemsize) {
	int c = 0;
	struct linkedlistitem* i = (struct linkedlistitem*)*ptr;
	if (i == NULL) {
		return linkedlist_Add(ptr,item,itemsize);
	}
	while (i->next != NULL) {
		i = i->next;c++;
	}
	if (c <= 0) {
		return linkedlist_Add(ptr,item,itemsize);
	}else{
		return linkedlist_Add((void**)&i,item,itemsize);
	}
}
void linkedlist_FreeItem(void* item) {
	if (item == NULL) {return;}
	struct linkedlistitem* i = (struct linkedlistitem*)item;
	if (i->data != NULL) {free(i->data);}
	free(i);
}
void linkedlist_RemoveItem(void** list, void* item) {
	struct linkedlistitem* i = (struct linkedlistitem*)item;
	if (i->prev == NULL) {
		//we got the first item
		if (i->next != NULL) {
			if (LLISTDEBUG == 1) {
				printf("linkedlist: first item of a longer list deleted -> change list pointer\n");
			}
			//ok, change list pointer as we remove the first item
			*list = i->next;
			i->next->prev = NULL;
			linkedlist_FreeItem(i);
			return;
		}else{
			if (LLISTDEBUG == 1) {
				printf("linkedlist: first and only item deleted -> trash list\n");
			}
			//it is the last one, so free it
			linkedlist_FreeItem(i);
			*list = NULL;
			return;
		}
	}else{
		if (LLISTDEBUG == 1) {
			printf("linkedlist: deleting some item in between\n");
		}
		if (i->next != NULL) {
			i->next->prev = i->prev;
		}
		i->prev->next = i->next;
		linkedlist_FreeItem(i);
		return;
	}
}
void* linkedlist_NextItem(void* item) {
	return ((struct linkedlistitem*)item)->next;
}
void* linkedlist_GetData(void* item) {
	return (struct linkedlistitem*)((struct linkedlistitem*)item)->data;
}
void linkedlist_SetData(void* item, void* ptr) {
	((struct linkedlistitem*)item)->data = ptr;
}
void linkedlist_Free(void** ptr) {
	if (ptr == NULL) {return;}
	void* item = *ptr;
	if (item == NULL) {return;}
	struct linkedlistitem* i = (struct linkedlistitem*)item;
	if (i->data != NULL) {free(i->data);}
	while (i->next != NULL) {
		i = i->next;
		free(i->prev);
		if (i->data != NULL) {free(i->data);}
	}
	free(i);
}

/* helper functions for lists with simple data contents */

int stringlist_CheckItem(void* list, const char* str) {
	void* s = list;
	int c = -1;
	while (s != NULL) {
		c++;
		if (linkedlist_GetData(s) != NULL) {
			if (strcasecmp(linkedlist_GetData(s),str) == 0) {
				return c;
			}
		}
		s = linkedlist_NextItem(s);
	}
	return -1;
}
int intlist_CheckItem(void* list, int number) {
	void* s = list;
	int* i;
	while (s != NULL) {
		i = (int*)linkedlist_GetData(s);
		if (i != NULL) {
			if (*i == number) {return 0;}
		}
		s = linkedlist_NextItem(s);
	}
	return -1;
}
#ifndef NONATIVEWEIRCD
int masklist_CheckItem(void* list, const char* str) {
	void* s = list;
	int c = -1;
	while (s != NULL) {
		c++;
		if (linkedlist_GetData(s) != NULL) {
			if (masks_Match(str,linkedlist_GetData(s)) == 0) {
				return c;
			}
		}
		s = linkedlist_NextItem(s);
	}
	return -1;
}
#endif
void** linkedlist_ArrayFromList(void* list) {
	int count = 0;
	void* item = list;
	while (item != NULL) {
		count++;
		item = linkedlist_NextItem(item);
	}
	char** arrayspace = (char**)malloc(sizeof(void*)*count);
	count = 0;
	item = list;
	while (item != NULL) {
		arrayspace[count] = linkedlist_GetData(item);
		count++;
		item = linkedlist_NextItem(item);
	}
	return (void**)arrayspace;
}