//    WeIRCd IRC server software - sub file
//    Copyright (C) 2007-2009 Jonas Thiem
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License v3 as published
//    by the Free Software Foundation.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

/* modulesmanager.c - Simple managed module loader */
/* --------------------- */
/* Loads libs/modules or unloads them. See modulesmanager.h for long description. */

#include "weircd.h"
#ifndef WIN
#include <dlfcn.h>
#endif

struct modlistitem { // a modules list entry of the modules manager's list. keeps important information
	char* modfilename;
	char* modinfo_name;
	char* modinfo_version;
	void* ptr;
	int isweircdmodule; //if not 1 then this is an external lib loaded by some module
};
void* modlist = NULL; // our linked list of loaded modules

const char* modman_GetNameByPtr(void* ptr) {
	/* Simply get the linked list data of the list entry represented by ptr
	   and return its name property */
	return ((struct modlistitem*)linkedlist_GetData(ptr))->modfilename;
}
void* modman_GetPtrByName(const char* modname) {
	/* Cycle through all list entries and compare name */
	void* s = modlist;
	struct modlistitem* i;
	while (s != NULL) {//cycle to the end of the list
		i = (struct modlistitem*)linkedlist_GetData(s);
		/* Check if name of data struct of that item matches: */
		if (strcasecmp(i->modfilename,modname) == 0) {
			return s; //it does, return the pointer to it
		}
		s = linkedlist_NextItem(s); //go to next item
	}
	return NULL; //no match found
}
static void _unloadmod(void* ptr) {
	/* Unload a specific module by module pointer (NOT list entry pointer) */
	
	/* OS specific code: */
#ifndef WIN
	dlclose(ptr);
#endif
#ifdef WIN
	FreeLibrary((HMODULE*)ptr);
#endif
}
static int _loadmod(void** ptr, const char* path) {
	/* Load a specific module and return a module pointer (NOT list entry pointer) */
	
	/* We need to do the actual loading OS specific */
	
	#ifndef WIN
	/* UNIX code to load our .so
	   Saves the returned pointer into the provided
	   pointer address */
	*ptr = dlopen(path,RTLD_NOW|RTLD_GLOBAL);
	/*if (*ptr == NULL) { //debugging stuff we sometimes need
		printf("modloadfail: %u|%s\n",*ptr,dlerror());
	}*/
	#endif
	
	#ifdef WIN
	/* Windows code to load our .dll.
	   Saves the returned pointer into the provided
	   pointer address */
	*ptr = (void*)LoadLibrary(path);
	/*if (*ptr == NULL) { //debugging stuff
		int i = GetLastError();
		printf("error:%d\n",i);
	}*/
	#endif
	if (*ptr == NULL) {return -1;}//NULL pointer? -> loading failed
	/* Apparently it worked, return 0 */
	return 0;
}
int modman_Load(const char* modname) {
	if (modman_GetPtrByName(modname) != NULL) {return 0;} //module is already loaded
	
	/* First construct a full path of the name including a valid file extension */
	char tgpath[300];
	#ifndef WIN
	snprintf(tgpath,299,"%s/%s%s",moduledir,modname,modext);tgpath[299] = 0;
	#endif
	#ifdef WIN
	snprintf(tgpath,299,"%s\\%s%s",moduledir,modname,modext);tgpath[299] = 0;
	#endif
	
	/* Now get a module list entry struct, zero it and copy the name into it */
	struct modlistitem moddat;
	memset(&moddat,0,sizeof(moddat));
	moddat.modfilename = malloc(strlen(modname)+1);
	if (moddat.modfilename == NULL) {
		/* Out of memory or something? */
		return -1;
	}
	strcpy(moddat.modfilename,modname);
	
	/* Attempt to load the module. Will write module pointer into moddat.ptr on success */
	if (_loadmod(&moddat.ptr,tgpath) != 0) {
		#ifndef WIN
		printf("dlerror() for \"%s\": %s\n",tgpath,dlerror());
		#endif
		/* Failed - free file name string and return -1 */
		free(moddat.modfilename);
		return -1;
	}
	
	/* We managed to load it - now add the module list entry holding the name and the pointer to our list */
	void* s = linkedlist_Append(&modlist,&moddat,sizeof(moddat));
	if (s == NULL) {
		/* We weren't able to add it :-( - unload the module and the file name string again and return -1 */
		_unloadmod(moddat.ptr);
		free(moddat.modfilename);
		return -1;
	}
	
	/* Done! */
	printf(">> %s loaded\n",moddat.modfilename);
	return 0;	
}
void modman_UnloadByPtr(void* ptr) {	
	/* Unload a module by its module list entry pointer */
	
	/* First, obtain the struct which the item contains: */
	struct modlistitem* p = (struct modlistitem*)linkedlist_GetData(ptr);
	if (p == NULL) {return;} //not good
	
	/* Unload the module */
	_unloadmod(p->ptr);
	printf(">> %s now unloaded\n",p->modfilename); //debug message
	
	/* Free the file name string */
	free(p->modfilename);
	if (p->modinfo_name != NULL) {
		/* If present, free the name/title string */
		free(p->modinfo_name);
	}
	/* If present, free the version info string */
	if (p->modinfo_version != NULL) {free(p->modinfo_version);}
	
	/* Finally, remove the now empty item from the list */
	linkedlist_RemoveItem(&modlist,(void*)ptr);
}
void modman_Unload(const char* modname) {
	/* Unload a module by name. Simply obtain the pointer and call the
	  UnloadByPtr routine :-) */
	void* item = modman_GetPtrByName(modname);
	if (item == NULL) {return;} //the module doesn't exist
	modman_UnloadByPtr(item);
}
void* modman_GetDataByPtr(void* modptr, const char* data) {
	/* Get the contained struct of the module list entry */
	struct modlistitem* p = (struct modlistitem*)linkedlist_GetData(modptr);
	if (p == NULL) {return NULL;} //shouldn't happen with a valid pointer
		
	/* Again OS dependant code to resolve the symbol */
#ifndef WIN
	return dlsym(p->ptr,data);
#endif
#ifdef WIN
	return (void*)GetProcAddress((HMODULE)p->ptr,data);
#endif
}
void* modman_GetData(const char* modname, const char* data) {
	/* Resolve a symbol and return the resulting pointer */
	
	/* Obtain the module list entry of that module */
	void* item = modman_GetPtrByName(modname);
	if (item == NULL) {return NULL;} //the module isn't loaded!
		
	/* Get data by that ptr*/
	return modman_GetDataByPtr(item, data);
}